#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpphik/hppHikLinearSystem.hh"

using namespace boost::numeric::ublas;

ChppHikLinearSystem::ChppHikLinearSystem ( unsigned int inNparam )
{
    attMatrix.resize ( 0, inNparam );
}

unsigned int ChppHikLinearSystem::dimension() const
{
    return attValue.size();
}

bool ChppHikLinearSystem::expand ( const matrixNxP& inA, const vectorN& inb, std::vector<unsigned int>* inIndices )
{
    if ( inA.size2() != attMatrix.size2() )
    {
        std::cout << "ChppHikLinearSystem::addConstraint() dimension mismatch\n";
        return false;
    }

    unsigned int start = attMatrix.size1();;
    if ( !inIndices )
    {
        attMatrix.resize ( start+inA.size1(),attMatrix.size2(),true );
        subrange ( attMatrix,start,attMatrix.size1(),0,attMatrix.size2() ) = inA;
        attValue.resize ( start+inb.size(),true );
        subrange ( attValue,start,attValue.size() ) = inb;
    }
    else
    {
        unsigned int dim = inb.size();
        for ( unsigned int i=0; i < inIndices->size();i++ )
        {
            if ( ( *inIndices ) [i] >= dim )
            {
                std::cout << "ChppHikLinearSystem::addConstraint() index out of bounds\n";
                return false;
            }
        }
        attMatrix.resize ( start+inIndices->size(),attMatrix.size2(),true );
        attValue.resize ( start+inIndices->size(),true );
        for ( unsigned int i=0; i < inIndices->size();i++ )
        {
            row ( attMatrix, start+i ) =  row ( inA, ( *inIndices ) [i] );
            attValue ( start+i ) =  inb ( ( *inIndices ) [i] );
        }
    }
    return true;
}

bool ChppHikLinearSystem::disableColumns ( const vectorN& inMask )
{
    if ( inMask.size() != attMatrix.size2() )
        return false;
    zero_vector<double> zeros ( attMatrix.size1() );

    for ( unsigned int i=0;i<inMask.size();i++ )
        if ( inMask ( i ) ==0 )
            column ( attMatrix, i ) = zeros;

    return true;
}

void ChppHikLinearSystem::clear()
{
    attMatrix.resize ( 0,attMatrix.size2(),false );
    attValue.resize ( 0,false );
}

const vectorN& ChppHikLinearSystem::value()
{
    return attValue;
}

const matrixNxP& ChppHikLinearSystem::jacobian()
{
    return attMatrix;
}


matrixNxP& ChppHikLinearSystem::matrix()
{
    return attMatrix;
}

vectorN& ChppHikLinearSystem::vector()
{
    return attValue;
}
