
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/io.hpp"
#include "hpphik/hppHikTools.hh"


#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J
#define V3_I  MAL_S3_VECTOR_ACCESS

double ChppHikTools::attEps = 1e-8;


void ChppHikTools::splitM4(const matrix4d& inM, matrix3d& outR, vector3d& outV)
{
    unsigned int i,j;
    for (i=0;i<3;i++)
    {
        for (j=0;j<3;j++)
            M3_IJ(outR,i,j) = M4_IJ(inM,i,j);
        V3_I(outV,i) = M4_IJ(inM,i,3);
    }
}

bool ChppHikTools::Matrix4dFromVec(const vectorN& inVec, matrix4d& outH)
{
    if (inVec.size()<6)
        return false;

    vector3d NE_wn;
    V3_I(NE_wn,0) = inVec(3);
    V3_I(NE_wn,1) = inVec(4);
    V3_I(NE_wn,2) = inVec(5);
    
    double th = MAL_S3_VECTOR_NORM(NE_wn);
    
    if (th< attEps)
    {
        MAL_S4x4_MATRIX_SET_IDENTITY(outH);
    }
    else
    {
        NE_wn = NE_wn / th;
        double ct = cos(th);
        double lct= (1-ct);
        double st = sin(th);
        M4_IJ(outH,0,0) = ct + NE_wn[0]*NE_wn[0]* lct;
        M4_IJ(outH,0,1) = NE_wn[0]*NE_wn[1]*lct-NE_wn[2]*st;
        M4_IJ(outH,0,2) = NE_wn[1] * st+NE_wn[0]*NE_wn[2]*lct;
        M4_IJ(outH,1,0) = NE_wn[2]*st +NE_wn[0]*NE_wn[1]*lct;
        M4_IJ(outH,1,1) = ct + NE_wn[1]*NE_wn[1]*lct;
        M4_IJ(outH,1,2) = -NE_wn[0]*st+NE_wn[1]*NE_wn[2]*lct;
        M4_IJ(outH,2,0) = -NE_wn[1]*st+NE_wn[0]*NE_wn[2]*lct;
        M4_IJ(outH,2,1) = NE_wn[0]*st + NE_wn[1]*NE_wn[2]*lct;
        M4_IJ(outH,2,2) = ct + NE_wn[2]*NE_wn[2]*lct;
    }

    M4_IJ(outH,0,3) = inVec(0);
    M4_IJ(outH,1,3) = inVec(1);
    M4_IJ(outH,2,3) = inVec(2);
    
    return true;
}

void ChppHikTools::M3toEulerZYX(matrix3d& inRot, vector3d& outEuler)
{
    double qz = atan2(M3_IJ(inRot,1,0),M3_IJ(inRot,0,0));
    double cz = cos(qz);
    double sz = sin(qz);
    double qx = atan2(M3_IJ(inRot,0,2)*sz - M3_IJ(inRot,1,2)*cz, -M3_IJ(inRot,0,1)*sz + M3_IJ(inRot,1,1)*cz);

    if (fabs(qx) > M_PI/2)
    {
        int sig = (qx>0)? 1 : -1;
        qz = qz - sig*M_PI;
        qx = qx - sig*M_PI;
        cz = -cz;
        sz = -sz;
    }
    double qy = atan2(-M3_IJ(inRot,2,0),M3_IJ(inRot,0,0)*cz + M3_IJ(inRot,1,0)*sz);

    outEuler[0] = qx;
    outEuler[1] = qy;
    outEuler[2] = qz;
}
