/*
 *  Copyright
 */

//#define CHRONO

#include <iostream>
#include "hppHikTest.h"
#include "hppGik/robot/hppRobotMotion.h"
#include "hppGik/constraints/hppGikPositionConstraint.h"
#include "hppGik/constraints/hppGikPlaneConstraint.h"
#include "hppGik/constraints/hppGikComConstraint.h"
#include "hppGik/constraints/hppGikTransformationConstraint.h"
#include "hppGik/core/hppGikMotionPlanElement.h"

#ifdef CHRONO
#include <time.h>
#include <sys/time.h>
#endif

#include "MatrixAbstractLayer/MatrixAbstractLayer.h"
#include "robotDynamics/jrlRobotDynamicsObjectConstructor.h"
#include "dynamicsJRLJapan/HumanoidDynamicMultiBody.h"
#include "hrp2Dynamics/hrp2OptHumanoidDynamicRobot.h"
#include "hppGik/hppGikTools.h"
#include "hpphik/hppHikSolver.hh"

using namespace boost::numeric::ublas;

ChppHikTest::ChppHikTest() : attSamplingPeriod ( 5e-3 )
{
    createHumanoidRobot();
    attDs.resize ( attRobot->numberDof()-6 );
    attUb.resize ( attRobot->numberDof()-6 );
    attLb.resize ( attRobot->numberDof()-6 );
    att_inv_Di_Ds.resize ( attRobot->numberDof()-6 );


    attMaxVelo = 0.002;

    double DiRatio = 0.10;
    double DiDsRatio = 0.05;

    double ul,ll,di;
    for ( unsigned int i =0; i<attDs.size();i++ )
    {
        ul = attRobot->upperBoundDof ( i +6 );
        ll = attRobot->lowerBoundDof ( i +6 );
        di = DiRatio* ( ul-ll );
        attDs ( i ) = DiDsRatio*di;
        att_inv_Di_Ds ( i ) = 1/ ( di-attDs ( i ) );
    }
}

ChppHikTest::~ChppHikTest()
{
    delete attStandingRobot;
}


void obstdist ( vectorN& p1, vectorN& p2, vectorN& p, vectorN& n, double& d )
{
    vectorN vec = p2 - p1;
    double l = norm_2 ( vec );
    vectorN u = vec/l;

    vectorN p1p = p - p1;
    double x = inner_prod ( p1p,u );

    if ( x>=l )
    {
        n = p - p2;
        d = norm_2 ( n );
        n = n/d;
    }
    else if ( x<=0 )
    {
        d = norm_2 ( p1p );
        n = p1p/d;
    }
    else
        if ( ( p1p ( 0 ) *u ( 1 ) - p1p ( 1 ) *u ( 0 ) ) >0 )
        {
            n ( 0 ) = u ( 1 );
            n ( 1 ) = -u ( 0 );
        }
        else
        {
            n ( 0 ) = -u ( 1 );
            n ( 1 ) = u ( 0 );
        }

    d = inner_prod ( n,p1p );
}

void buildCollisionTask ( double maxspeed, ChppHikLinearSystem& task, vectorN& q, bool&  changed)
{
    /*
    vectorN p1 ( 2 );
    p1 ( 0 ) = -1;
    p1 ( 1 ) = 1;
    vectorN p2 ( 2 );
    p2 ( 0 ) = 2;
    p2 ( 1 ) = 1;

    double ds = 0.1;
    double di = 0.4;
    double d = 0.0;

    changed = false;
    vectorN n ( 2 );

    obstdist ( p1, p2, q, n, d);

    if ( d <= di )
    {
        changed = true;
        task.vector() ( 0 ) = maxspeed* ( d-ds ) / (di-ds );
        task.matrix() ( 0,0 ) =  -n ( 0 );
        task.matrix() ( 0,1 ) =  -n ( 1 );
    }
    */
    changed  = false;
    double ds = 0.1;
    double di = 0.4;
    double Xobs = 1.0;
    double Xs = Xobs + ds;
    double Xi = Xobs + di;
    if ( q(1) <= Xi )
    {
        task.vector() ( 0 ) = maxspeed * ( q(1)-Xs ) / (Xi - Xs );
        task.matrix() ( 0,0 ) = 0;
        task.matrix() ( 0,1 ) = -1;
        changed = true;
    }

}

bool ChppHikTest::solveSimple()
{
    ChppHikSolverBasic solver ( 2 );
    ChppHikLinearSystem collisionTask ( 2 );
    collisionTask.matrix().resize ( 1,2 );
    collisionTask.vector().resize ( 1 );
    ChppHikLinearSystem reachTask ( 2 );
    ChppHikLinearSystem preferredRegionTask ( 2 );
    preferredRegionTask.matrix().resize ( 1,2 );
    preferredRegionTask.vector().resize ( 1 );
    preferredRegionTask.matrix() ( 0,0 ) = 0;
    preferredRegionTask.matrix() ( 0,1 ) = -1;
    preferredRegionTask.vector() ( 0 ) = -1;

    vectorN q ( 2 );//[3;2.5];//[2;1.5]//[3;2.5];//
    q ( 0 ) = 2;
    q ( 1 ) = 1.5;

    vectorN dq ( 2 );
    dq ( 0 ) = 1;
    dq ( 1 ) = 1;

    double maxspeed = 0.01;
    vectorN target ( 2 ), steptarget ( 2 );
    target ( 0 ) = 0;
    target ( 1 ) = -0.1;

    double dqmax = 0.02;
    vectorN ub = scalar_vector<double> ( 2,dqmax );
    vectorN lb = -ub;

    std::vector< std::pair<ChppHikLinearSystem*,bool> > tasks;

    bool changed = false;
    unsigned int i,b;
    while ( 1 )
    {
        tasks.clear();

        //Reach
        steptarget = target - q;
        double normstep = norm_2 ( steptarget );

        if ( normstep >= 4*maxspeed )
        {
            reachTask.matrix().resize ( 1,2 );
            reachTask.vector().resize ( 1 );
            row ( reachTask.matrix(),0 ) = steptarget/normstep;
            reachTask.vector() ( 0 ) = maxspeed;
        }
        else
            if ( normstep > 1e-3 )
            {
                reachTask.matrix() = identity_matrix<double> ( 2 );
                reachTask.vector() = steptarget;
            }
            else
            {
                std::cout << "Goal reached" << std::endl;
                break; // end, we reached the goal
            }

        tasks.push_back ( std::make_pair(&reachTask,true) );

        //Collision avoidance if needed
        buildCollisionTask ( maxspeed, collisionTask, q, changed );
        if ( changed )
                tasks.push_back ( std::make_pair(&collisionTask,false) );
        
        //tasks.push_back ( std::make_pair(&preferredRegionTask,false) );

        //Set parameter bounds
        solver.clearBounds();
        for ( b=0;b<2;b++ )
        {
            solver.upperBound ( b, ub( b ) );
            solver.lowerBound ( b, lb( b ) );
        }
        solver.initFeasibleSet();
        dq.clear();
        for ( i = 0; i< tasks.size(); i++ )
        {
            solver.clearTasks();
            solver.addTask ( *tasks[i].first, tasks[i].second );
            int retV = solver.solve ( dq, 0.1 );
            if ( retV !=0 )
            {
                std::cout << "retV = " << retV << " for task ranked "<< i<< std::endl;
                return false;
            }
/*
            std::cout << "tasks["<< i <<"]" << std::endl;
            std::cout << tasks[i].first->matrix() << std::endl;
            std::cout << tasks[i].first->vector() << std::endl;
       */     
            if ( i!= tasks.size()-1 )
                solver.updateFeasibleSet();
        }
        dq = solver.solution();
        
        if ( norm_2 ( dq ) < 1e-3 )
        {
            std::cout << "almost not moving" << std::endl;
            break;
        }

        
        q = q + dq;
        std::cout << q(0) << " " << q(1) << std::endl;

    }
    return true;
}




bool ChppHikTest::solve()
{
    ChppRobotMotion attSolutionMotion ( attRobot,0.0,attSamplingPeriod );

    ChppHikSolver hikSolver ( *attRobot );
    hikSolver.rootJoint ( * ( attRobot->rightFoot() ) );


    std::vector<CjrlGikStateConstraint*> tasks;
    vector3d p, localPoint;

    vector3d com = attRobot->positionCenterOfMass();
    ChppGikComConstraint comc ( *attRobot,com[0],com[1] );
    tasks.push_back ( &comc );

    matrix4d LFT=  attRobot->leftFoot()->currentTransformation();
    CjrlJoint& lfJ = * ( attRobot->leftFoot() );
    p[0] = p[1] = p[2] = 0;
    ChppGikTransformationConstraint footc ( *attRobot,lfJ,p,LFT );
    tasks.push_back ( &footc );

    matrix4d curT=  attRobot->rightWrist()->currentTransformation();
    CjrlJoint& rwJoint = * ( attRobot->rightWrist() );
    localPoint = attRobot->rightHand()->centerInWristFrame();
    p = curT*localPoint;
    ChppGikPositionConstraint pc ( *attRobot,rwJoint,localPoint, p );
    tasks.push_back ( &pc );

    vectorN solutionConfig, veloSolution ( attRobot->numberDof() );
    int retV;

    for ( unsigned int b=0;b<attUb.size();b++ )
    {
        hikSolver.defaultUpperBound ( b, attMaxVelo );
        hikSolver.defaultLowerBound ( b, -attMaxVelo );
    }

    vectorN outEvaluation;
    vectorN mask = attStandingRobot->maskFactory()->wholeBodyMask();
    for ( unsigned int j = 0; j< 6;j++ )
        mask ( j ) = 0.0;
    hikSolver.setMask ( mask );

    for ( unsigned int j = 0; j< 500;j++ )
    {
        p = pc.worldTarget();
        p[0] += 0.001;
        pc.worldTarget ( p );

        updateBounds();
        hikSolver.clearBounds();
        for ( unsigned int b=0;b<attUb.size();b++ )
        {
            hikSolver.upperBound ( b, attUb ( b ) );
            hikSolver.lowerBound ( b, attLb ( b ) );
        }

        hikSolver.initFeasibleSet();

        veloSolution.clear();
        for ( unsigned int t = 0; t<tasks.size();t++ )
        {
            tasks[t]->jacobianRoot ( * ( attRobot->rightFoot() ) );
            tasks[t]->computeValue();
            tasks[t]->computeJacobian();
        }
#ifdef CHRONO
        struct timeval* Tps = ( struct timeval* ) malloc ( sizeof ( struct timeval ) );
        struct timeval* Tpf = ( struct timeval* ) malloc ( sizeof ( struct timeval ) );
        struct timezone* Tzp = 0;

        gettimeofday ( Tps, Tzp );
#endif
        for ( unsigned int t = 0; t<tasks.size();t++ )
        {

            hikSolver.clearTasks();
            hikSolver.addTask ( *tasks[t],true );
            retV = hikSolver.solve ( veloSolution );
            if ( retV!=0 )
            {
                std::cout << "Could not solve Hik problem at iteration "<<j<<" error code = " << retV<<"\n";
                attSolutionMotion.dumpTo ( "hikmotion" );
                return false;
            }
            veloSolution = hikSolver.solution();
            if ( t != tasks.size()-1 )
                hikSolver.updateFeasibleSet();
        }
#ifdef CHRONO
        gettimeofday ( Tpf, Tzp );
        printf ( "Iteration //ld solved in: //ld ms\n", j, ( Tpf->tv_sec-Tps->tv_sec ) *1000 + ( Tpf->tv_usec-Tps->tv_usec ) /1000 );
#endif


        hikSolver.computeSolutionConfiguration ( veloSolution );
        solutionConfig = hikSolver.solutionConfiguration();
        attRobot->currentConfiguration ( solutionConfig );
        attRobot->computeForwardKinematics();
        attSolutionMotion.appendSample ( solutionConfig,p, p, p, p );
    }
    attSolutionMotion.dumpTo ( "hikmotion" );
    return true;
}

void ChppHikTest::updateBounds()
{
    double ul,ll,dp,dn;
    for ( unsigned int i =0; i<attDs.size();i++ )
    {
        ul = attRobot->upperBoundDof ( i+6 );
        ll = attRobot->lowerBoundDof ( i+6 );
        dp = ul - attRobot->currentConfiguration() ( i+6 );
        attUb ( i ) = attMaxVelo* ( dp-attDs ( i ) ) *att_inv_Di_Ds ( i );
        dn = attRobot->currentConfiguration() ( i+6 )-ll;
        attLb ( i ) = -attMaxVelo* ( dn-attDs ( i ) ) *att_inv_Di_Ds ( i );
    }
}

void ChppHikTest::createHumanoidRobot()
{
    CjrlRobotDynamicsObjectConstructor<
    dynamicsJRLJapan::DynamicMultiBody,
    Chrp2OptHumanoidDynamicRobot,
    dynamicsJRLJapan::JointFreeflyer,
    dynamicsJRLJapan::JointRotation,
    dynamicsJRLJapan::JointTranslation,
    dynamicsJRLJapan::Body> jrlRobotFactory;

    attRobot = jrlRobotFactory.createhumanoidDynamicRobot();

    dynamicsJRLJapan::HumanoidDynamicMultiBody *aHDMB;
    aHDMB = ( dynamicsJRLJapan::HumanoidDynamicMultiBody* ) attRobot;

    std::string path = "./";
    std::string name = "HRP2.wrl";
    aHDMB->parserVRML ( path,name,"./HRP2LinkJointRank.xml" );
    std::string aName="./HRP2Specificities.xml";
    aHDMB->SetHumanoidSpecificitiesFile ( aName );
    aHDMB->SetTimeStep ( attSamplingPeriod );
    aHDMB->setComputeVelocity ( false );
    aHDMB->setComputeMomentum ( false );
    aHDMB->setComputeCoM ( true );
    aHDMB->setComputeAcceleration ( false );
    aHDMB->setComputeZMP ( false );

    aHDMB->setComputeSkewCoM ( false );
    aHDMB->setComputeAccelerationCoM ( false );
    aHDMB->setComputeBackwardDynamics ( false );


    unsigned int nDof = attRobot->numberDof();
    vectorN halfsittingConf ( nDof );

    //Half sitting
    double dInitPos[40] =
    {
        0.0, 0.0, -26.0, 50.0, -24.0, 0.0, // right leg

        0.0, 0.0, -26.0, 50.0, -24.0, 0.0, // left leg

        0.0, 0.0, // chest

        0.0, 0.0, // head

        15.0, -10.0, 0.0, -30.0, 0.0, 0.0, // right arm

        10.0, // right hand clench

        15.0,  10.0, 0.0, -30.0, 0.0, 0.0, // left arm

        10.0, // left hand clench

        -10.0, 10.0, -10.0, 10.0, -10.0, // right hand parallel mechanism
        -10.0, 10.0, -10.0, 10.0, -10.0  // left hand parallel mechanism
    };

    //waist x y z
    halfsittingConf ( 0 ) = 0.0;
    halfsittingConf ( 1 ) = 0.0;
    halfsittingConf ( 2 ) = 0.6487;
    //waist roll pitch yaw
    halfsittingConf ( 3 ) = 0.0;
    halfsittingConf ( 4 ) = 0.0;
    halfsittingConf ( 5 ) = 0.0;

    //joints
    for ( unsigned int i=6;i<nDof;i++ )
        halfsittingConf ( i ) = dInitPos[i-6]*M_PI/180;

    zero_vector<double> zeros ( attRobot->numberDof() );
    attRobot->currentConfiguration ( halfsittingConf );
    attRobot->currentVelocity ( zeros );
    attRobot->currentAcceleration ( zeros );
    attRobot->computeForwardKinematics();

    //set gaze origin and direction
    vector3d gazeDir,gazeOrigin;
    gazeDir[0] = 1;
    gazeDir[1] = 0;
    gazeDir[2] = 0;

    gazeOrigin[0] = 0;
    gazeOrigin[1] = 0;
    gazeOrigin[2] = 0.118;

    aHDMB->gaze ( ( const vector3d& ) gazeDir, ( const vector3d& ) gazeOrigin );

    
    ChppGik2DShape leftFootShape, rightFootShape;

    ChppGik2DVertex vert;

    vert.x = 0.135;
    vert.y = 0.079;
    leftFootShape.vertices.push_back ( vert );
    vert.x = -0.105;
    leftFootShape.vertices.push_back ( vert );
    vert.y = -0.059;
    leftFootShape.vertices.push_back ( vert );
    vert.x = 0.135;
    leftFootShape.vertices.push_back ( vert );

    vert.x = 0.135;
    vert.y = 0.059;
    rightFootShape.vertices.push_back ( vert );
    vert.x = -0.105;
    rightFootShape.vertices.push_back ( vert );
    vert.y = -0.079;
    rightFootShape.vertices.push_back ( vert );
    vert.x = 0.135;
    rightFootShape.vertices.push_back ( vert );

    attStandingRobot = new ChppGikStandingRobot ( *attRobot, leftFootShape, rightFootShape );

    attStandingRobot->staticState ( halfsittingConf );
}
